#include "loc.h"
#include "logger.h"
#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <token.h>
#include <dynarray.h>
#include <parser/parser.h>
#include <parser/precomp.h>
#include <parser/tokcmp.h>
#include <parser/ast.h>

// compiles values in the token list, will error if invalid tokens
int compile_value(token_t* tokens) {
    // 'f' for float, 'i' for int, 's' for string, 'c' for char
    char exclusive_type = '\0';
    token_t* tmp_toks = dynarray_create(token_t);
    token_t tok = {0};
    parser_state_t ps = {0};
    ps.tokens = tokens;
    while (dynarray_pop(tokens, &tok) == 0) {
        switch (tok.type) {
        case (TT_PUSH_CSTR):
        case (TT_PUSH_STR): {
            // TODO: Add C style string literall joining
            if (dynarray_length(tokens) != 0) {
                log_error(&tok.loc, "Constants do not support C style string literal joining yet");
                return 1;
            }
            dynarray_push(tmp_toks, tok);
        }; break;
        case (TT_PUSH_INT):
        case (TT_PUSH_FLOAT):
        case (TT_PUSH_CHAR):
        case (TT_PUSH_BOOL):
        case (TT_PUSH_MEM): {
            dynarray_push(tmp_toks, tok);
        }; break;
        case (TT_KW): {
            // TODO: Support if statements in constants at compile time against other constants
            log_error(&tok.loc, "Keywords are currently not supported in constants");
            return 1;
        }
        case (TT_IDENT): {
            // TODO: Support having other constants as arguments in constants, and memory values for making offsets
            log_error(&tok.loc, "Identifiers are currently not supported in constants");
            return 1;
        }
        case (TT_OP): {
            switch (tok.op_type) {
            case (OP_ADD): {
                token_t* t2 = &tmp_toks[dynarray_length(tmp_toks - 1)];
                token_t* t1 = &tmp_toks[dynarray_length(tmp_toks - 2)];
                if (t1->type == TT_PUSH_INT && t2->type == TT_PUSH_INT) {
                    size_t val = t1->int_v + t2->int_v;
                    loc_t loc = t1->loc;
                    dynarray_pop(tokens, NULL);
                    dynarray_pop(tokens, NULL);
                    token_t tok = {.type = TT_PUSH_INT, .int_v = val, .loc = loc};
                    dynarray_push(tokens, tok);

                } else if (t1->type == TT_PUSH_FLOAT && t2->type == TT_PUSH_FLOAT) {
                    double_t val = t1->float_v + t2->float_v;
                    loc_t loc = t1->loc;
                    dynarray_pop(tokens, NULL);
                    dynarray_pop(tokens, NULL);
                    token_t tok = {.type = TT_PUSH_FLOAT, .int_v = val, .loc = loc};
                    dynarray_push(tokens, tok);

                } else if (t1->type == TT_PUSH_MEM && t2->type == TT_PUSH_INT) {
                    size_t val = t1->int_v + t2->int_v;
                    loc_t loc = t1->loc;
                    dynarray_pop(tokens, NULL);
                    dynarray_pop(tokens, NULL);
                    token_t tok = {.type = TT_PUSH_INT, .int_v = val, .loc = loc};
                    dynarray_push(tokens, tok);
                } else {
                    const char* s1 = get_tok_str_dbg(t1);
                    const char* s2 = get_tok_str_dbg(t2);
                    log_error(&t1->loc, "Tokens %s and %s cannot be added together", s1, s2);
                    return 1;
                }
            }; break;
            case (OP_SUB): {
                token_t* t2 = &tmp_toks[dynarray_length(tmp_toks - 1)];
                token_t* t1 = &tmp_toks[dynarray_length(tmp_toks - 2)];
                if (t1->type == TT_PUSH_INT && t2->type == TT_PUSH_INT) {
                    size_t val = t1->int_v - t2->int_v;
                    loc_t loc = t1->loc;
                    dynarray_pop(tokens, NULL);
                    dynarray_pop(tokens, NULL);
                    token_t tok = {.type = TT_PUSH_INT, .int_v = val, .loc = loc};
                    dynarray_push(tokens, tok);

                } else if (t1->type == TT_PUSH_FLOAT && t2->type == TT_PUSH_FLOAT) {
                    double_t val = t1->float_v - t2->float_v;
                    loc_t loc = t1->loc;
                    dynarray_pop(tokens, NULL);
                    dynarray_pop(tokens, NULL);
                    token_t tok = {.type = TT_PUSH_FLOAT, .int_v = val, .loc = loc};
                    dynarray_push(tokens, tok);

                } else if (t1->type == TT_PUSH_MEM && t2->type == TT_PUSH_INT) {
                    size_t val = t1->int_v - t2->int_v;
                    loc_t loc = t1->loc;
                    dynarray_pop(tokens, NULL);
                    dynarray_pop(tokens, NULL);
                    token_t tok = {.type = TT_PUSH_INT, .int_v = val, .loc = loc};
                    dynarray_push(tokens, tok);
                } else {
                    const char* s1 = get_tok_str_dbg(t1);
                    const char* s2 = get_tok_str_dbg(t2);
                    log_error(&t1->loc, "Tokens %s and %s cannot be subtracted from eachother", s1, s2);
                    return 1;
                }
            }; break;
            case (OP_MUL): {
                token_t* t2 = &tmp_toks[dynarray_length(tmp_toks - 1)];
                token_t* t1 = &tmp_toks[dynarray_length(tmp_toks - 2)];
                if (t1->type == TT_PUSH_INT && t2->type == TT_PUSH_INT) {
                    size_t val = t1->int_v * t2->int_v;
                    loc_t loc = t1->loc;
                    dynarray_pop(tokens, NULL);
                    dynarray_pop(tokens, NULL);
                    token_t tok = {.type = TT_PUSH_INT, .int_v = val, .loc = loc};
                    dynarray_push(tokens, tok);

                } else if (t1->type == TT_PUSH_FLOAT && t2->type == TT_PUSH_FLOAT) {
                    double_t val = t1->float_v * t2->float_v;
                    loc_t loc = t1->loc;
                    dynarray_pop(tokens, NULL);
                    dynarray_pop(tokens, NULL);
                    token_t tok = {.type = TT_PUSH_FLOAT, .int_v = val, .loc = loc};
                    dynarray_push(tokens, tok);

                } else if (t1->type == TT_PUSH_MEM && t2->type == TT_PUSH_INT) {
                    size_t val = t1->int_v * t2->int_v;
                    loc_t loc = t1->loc;
                    dynarray_pop(tokens, NULL);
                    dynarray_pop(tokens, NULL);
                    token_t tok = {.type = TT_PUSH_INT, .int_v = val, .loc = loc};
                    dynarray_push(tokens, tok);
                } else {
                    const char* s1 = get_tok_str_dbg(t1);
                    const char* s2 = get_tok_str_dbg(t2);
                    log_error(&t1->loc, "Tokens %s and %s cannot be multiplied together", s1, s2);
                    return 1;
                }
            }; break;
            case (OP_DIV): {
                token_t* t2 = &tmp_toks[dynarray_length(tmp_toks - 1)];
                token_t* t1 = &tmp_toks[dynarray_length(tmp_toks - 2)];
                if (t1->type == TT_PUSH_INT && t2->type == TT_PUSH_INT) {
                    size_t val = t1->int_v / t2->int_v;
                    loc_t loc = t1->loc;
                    dynarray_pop(tokens, NULL);
                    dynarray_pop(tokens, NULL);
                    token_t tok = {.type = TT_PUSH_INT, .int_v = val, .loc = loc};
                    dynarray_push(tokens, tok);

                } else if (t1->type == TT_PUSH_FLOAT && t2->type == TT_PUSH_FLOAT) {
                    double_t val = t1->float_v / t2->float_v;
                    loc_t loc = t1->loc;
                    dynarray_pop(tokens, NULL);
                    dynarray_pop(tokens, NULL);
                    token_t tok = {.type = TT_PUSH_FLOAT, .int_v = val, .loc = loc};
                    dynarray_push(tokens, tok);

                } else if (t1->type == TT_PUSH_MEM && t2->type == TT_PUSH_INT) {
                    size_t val = t1->int_v / t2->int_v;
                    loc_t loc = t1->loc;
                    dynarray_pop(tokens, NULL);
                    dynarray_pop(tokens, NULL);
                    token_t tok = {.type = TT_PUSH_INT, .int_v = val, .loc = loc};
                    dynarray_push(tokens, tok);
                } else {
                    const char* s1 = get_tok_str_dbg(t1);
                    const char* s2 = get_tok_str_dbg(t2);
                    log_error(&t1->loc, "Tokens %s and %s cannot be divided", s1, s2);
                    return 1;
                }
            }; break;
            case (OP_MOD): {
                token_t* t2 = &tmp_toks[dynarray_length(tmp_toks - 1)];
                token_t* t1 = &tmp_toks[dynarray_length(tmp_toks - 2)];
                if (t1->type == TT_PUSH_INT && t2->type == TT_PUSH_INT) {
                    size_t val = t1->int_v % t2->int_v;
                    loc_t loc = t1->loc;
                    dynarray_pop(tokens, NULL);
                    dynarray_pop(tokens, NULL);
                    token_t tok = {.type = TT_PUSH_INT, .int_v = val, .loc = loc};
                    dynarray_push(tokens, tok);
                } else if (t1->type == TT_PUSH_MEM && t2->type == TT_PUSH_INT) {
                    size_t val = t1->int_v % t2->int_v;
                    loc_t loc = t1->loc;
                    dynarray_pop(tokens, NULL);
                    dynarray_pop(tokens, NULL);
                    token_t tok = {.type = TT_PUSH_INT, .int_v = val, .loc = loc};
                    dynarray_push(tokens, tok);
                } else {
                    const char* s1 = get_tok_str_dbg(t1);
                    const char* s2 = get_tok_str_dbg(t2);
                    log_error(&t1->loc, "Tokens %s and %s cannot be modulo'd together", s1, s2);
                    return 1;
                }
            }; break;
            case (OP_EQ): {
                token_t* t2 = &tmp_toks[dynarray_length(tmp_toks - 1)];
                token_t* t1 = &tmp_toks[dynarray_length(tmp_toks - 2)];
                if (t1->type == t2->type) {
                    bool val = 0;
                    switch (t1->type) {
                    case (TT_PUSH_INT):
                        val = t1->int_v == t2->int_v;
                        break;
                    case (TT_PUSH_CSTR):
                    case (TT_PUSH_STR):
                        val = strcmp(t1->str_v, t2->str_v) == 0;
                        break;
                    case (TT_PUSH_FLOAT):
                        val = t1->float_v == t2->float_v;
                        break;
                    case (TT_PUSH_CHAR):
                        val = t1->char_v == t2->char_v;
                        break;
                    default:
                        const char* s1 = get_tok_str_dbg(t1);
                        const char* s2 = get_tok_str_dbg(t2);
                        log_error(&t1->loc, "Tokens %s and %s cannot be compared", s1, s2);
                        return 1;
                    }
                    dynarray_pop(tokens, NULL);
                    dynarray_pop(tokens, NULL);
                    token_t tok = {.type = TT_PUSH_BOOL, .int_v = val, .loc = t1->loc};
                    dynarray_push(tokens, tok);
                } else {
                    const char* s1 = get_tok_str_dbg(t1);
                    const char* s2 = get_tok_str_dbg(t2);
                    log_error(&t1->loc, "Tokens %s and %s cannot be compared", s1, s2);
                    return 1;
                }
            }; break;
            case (OP_GT): {
                token_t* t2 = &tmp_toks[dynarray_length(tmp_toks - 1)];
                token_t* t1 = &tmp_toks[dynarray_length(tmp_toks - 2)];
                if (t1->type == t2->type) {
                    bool val = 0;
                    switch (t1->type) {
                    case (TT_PUSH_INT):
                        val = t1->int_v > t2->int_v;
                        break;
                    case (TT_PUSH_FLOAT):
                        val = t1->float_v > t2->float_v;
                        break;
                    case (TT_PUSH_CHAR):
                        val = t1->char_v > t2->char_v;
                        break;
                    default:
                        const char* s1 = get_tok_str_dbg(t1);
                        const char* s2 = get_tok_str_dbg(t2);
                        log_error(&t1->loc, "Tokens %s and %s cannot be compared", s1, s2);
                        return 1;
                    }
                    dynarray_pop(tokens, NULL);
                    dynarray_pop(tokens, NULL);
                    token_t tok = {.type = TT_PUSH_BOOL, .int_v = val, .loc = t1->loc};
                    dynarray_push(tokens, tok);
                } else {
                    const char* s1 = get_tok_str_dbg(t1);
                    const char* s2 = get_tok_str_dbg(t2);
                    log_error(&t1->loc, "Tokens %s and %s cannot be compared", s1, s2);
                    return 1;
                }
            }; break;
            case (OP_LT): {
                token_t* t2 = &tmp_toks[dynarray_length(tmp_toks - 1)];
                token_t* t1 = &tmp_toks[dynarray_length(tmp_toks - 2)];
                if (t1->type == t2->type) {
                    bool val = 0;
                    switch (t1->type) {
                    case (TT_PUSH_INT):
                        val = t1->int_v < t2->int_v;
                        break;
                    case (TT_PUSH_FLOAT):
                        val = t1->float_v < t2->float_v;
                        break;
                    case (TT_PUSH_CHAR):
                        val = t1->char_v < t2->char_v;
                        break;
                    default:
                        const char* s1 = get_tok_str_dbg(t1);
                        const char* s2 = get_tok_str_dbg(t2);
                        log_error(&t1->loc, "Tokens %s and %s cannot be compared", s1, s2);
                        return 1;
                    }
                    dynarray_pop(tokens, NULL);
                    dynarray_pop(tokens, NULL);
                    token_t tok = {.type = TT_PUSH_BOOL, .int_v = val, .loc = t1->loc};
                    dynarray_push(tokens, tok);
                } else {
                    const char* s1 = get_tok_str_dbg(t1);
                    const char* s2 = get_tok_str_dbg(t2);
                    log_error(&t1->loc, "Tokens %s and %s cannot be compared", s1, s2);
                    return 1;
                }
            }; break;
            case (OP_GE): {
                token_t* t2 = &tmp_toks[dynarray_length(tmp_toks - 1)];
                token_t* t1 = &tmp_toks[dynarray_length(tmp_toks - 2)];
                if (t1->type == t2->type) {
                    bool val = 0;
                    switch (t1->type) {
                    case (TT_PUSH_INT):
                        val = t1->int_v >= t2->int_v;
                        break;
                    case (TT_PUSH_FLOAT):
                        val = t1->float_v >= t2->float_v;
                        break;
                    case (TT_PUSH_CHAR):
                        val = t1->char_v >= t2->char_v;
                        break;
                    default:
                        const char* s1 = get_tok_str_dbg(t1);
                        const char* s2 = get_tok_str_dbg(t2);
                        log_error(&t1->loc, "Tokens %s and %s cannot be compared", s1, s2);
                        return 1;
                    }
                    dynarray_pop(tokens, NULL);
                    dynarray_pop(tokens, NULL);
                    token_t tok = {.type = TT_PUSH_BOOL, .int_v = val, .loc = t1->loc};
                    dynarray_push(tokens, tok);
                } else {
                    const char* s1 = get_tok_str_dbg(t1);
                    const char* s2 = get_tok_str_dbg(t2);
                    log_error(&t1->loc, "Tokens %s and %s cannot be compared", s1, s2);
                    return 1;
                }
            }; break;
            case (OP_LE): {
                token_t* t2 = &tmp_toks[dynarray_length(tmp_toks - 1)];
                token_t* t1 = &tmp_toks[dynarray_length(tmp_toks - 2)];
                if (t1->type == t2->type) {
                    bool val = 0;
                    switch (t1->type) {
                    case (TT_PUSH_INT):
                        val = t1->int_v <= t2->int_v;
                        break;
                    case (TT_PUSH_FLOAT):
                        val = t1->float_v <= t2->float_v;
                        break;
                    case (TT_PUSH_CHAR):
                        val = t1->char_v <= t2->char_v;
                        break;
                    default:
                        const char* s1 = get_tok_str_dbg(t1);
                        const char* s2 = get_tok_str_dbg(t2);
                        log_error(&t1->loc, "Tokens %s and %s cannot be compared", s1, s2);
                        return 1;
                    }
                    dynarray_pop(tokens, NULL);
                    dynarray_pop(tokens, NULL);
                    token_t tok = {.type = TT_PUSH_BOOL, .int_v = val, .loc = t1->loc};
                    dynarray_push(tokens, tok);
                } else {
                    const char* s1 = get_tok_str_dbg(t1);
                    const char* s2 = get_tok_str_dbg(t2);
                    log_error(&t1->loc, "Tokens %s and %s cannot be compared", s1, s2);
                    return 1;
                }
            }; break;
            case (OP_NE): {
                token_t* t2 = &tmp_toks[dynarray_length(tmp_toks - 1)];
                token_t* t1 = &tmp_toks[dynarray_length(tmp_toks - 2)];
                if (t1->type == t2->type) {
                    bool val = 0;
                    switch (t1->type) {
                    case (TT_PUSH_INT):
                        val = t1->int_v != t2->int_v;
                        break;
                    case (TT_PUSH_CSTR):
                    case (TT_PUSH_STR):
                        val = strcmp(t1->str_v, t2->str_v) != 0;
                        break;
                    case (TT_PUSH_FLOAT):
                        val = t1->float_v != t2->float_v;
                        break;
                    case (TT_PUSH_CHAR):
                        val = t1->char_v != t2->char_v;
                        break;
                    default:
                        const char* s1 = get_tok_str_dbg(t1);
                        const char* s2 = get_tok_str_dbg(t2);
                        log_error(&t1->loc, "Tokens %s and %s cannot be compared", s1, s2);
                        return 1;
                    }
                    dynarray_pop(tokens, NULL);
                    dynarray_pop(tokens, NULL);
                    token_t tok = {.type = TT_PUSH_BOOL, .int_v = val, .loc = t1->loc};
                    dynarray_push(tokens, tok);
                } else {
                    const char* s1 = get_tok_str_dbg(t1);
                    const char* s2 = get_tok_str_dbg(t2);
                    log_error(&t1->loc, "Tokens %s and %s cannot be compared", s1, s2);
                    return 1;
                }
            }; break;
            case (OP_SHR): {
                token_t* t2 = &tmp_toks[dynarray_length(tmp_toks - 1)];
                token_t* t1 = &tmp_toks[dynarray_length(tmp_toks - 2)];
                if (t1->type == t2->type) {
                    size_t val = 0;
                    switch (t1->type) {
                    case (TT_PUSH_INT):
                        val = t1->int_v >> t2->int_v;
                        break;
                    default:
                        const char* s1 = get_tok_str_dbg(t1);
                        const char* s2 = get_tok_str_dbg(t2);
                        log_error(&t1->loc, "Tokens %s and %s cannot be SHR'd", s1, s2);
                        return 1;
                    }
                    dynarray_pop(tokens, NULL);
                    dynarray_pop(tokens, NULL);
                    token_t tok = {.type = TT_PUSH_INT, .int_v = val, .loc = t1->loc};
                    dynarray_push(tokens, tok);
                } else {
                    const char* s1 = get_tok_str_dbg(t1);
                    const char* s2 = get_tok_str_dbg(t2);
                    log_error(&t1->loc, "Tokens %s and %s cannot be SHR'd", s1, s2);
                    return 1;
                }
            }; break;
            case (OP_SHL): {
                token_t* t2 = &tmp_toks[dynarray_length(tmp_toks - 1)];
                token_t* t1 = &tmp_toks[dynarray_length(tmp_toks - 2)];
                if (t1->type == t2->type) {
                    size_t val = 0;
                    switch (t1->type) {
                    case (TT_PUSH_INT):
                        val = t1->int_v << t2->int_v;
                        break;
                    default:
                        const char* s1 = get_tok_str_dbg(t1);
                        const char* s2 = get_tok_str_dbg(t2);
                        log_error(&t1->loc, "Tokens %s and %s cannot be SHL'd", s1, s2);
                        return 1;
                    }
                    dynarray_pop(tokens, NULL);
                    dynarray_pop(tokens, NULL);
                    token_t tok = {.type = TT_PUSH_INT, .int_v = val, .loc = t1->loc};
                    dynarray_push(tokens, tok);
                } else {
                    const char* s1 = get_tok_str_dbg(t1);
                    const char* s2 = get_tok_str_dbg(t2);
                    log_error(&t1->loc, "Tokens %s and %s cannot be SHL'd", s1, s2);
                    return 1;
                }
            }; break;
            case (OP_BOR): {
                token_t* t2 = &tmp_toks[dynarray_length(tmp_toks - 1)];
                token_t* t1 = &tmp_toks[dynarray_length(tmp_toks - 2)];
                if (t1->type == t2->type) {
                    size_t val = 0;
                    switch (t1->type) {
                    case (TT_PUSH_INT):
                        val = t1->int_v | t2->int_v;
                        break;
                    default:
                        const char* s1 = get_tok_str_dbg(t1);
                        const char* s2 = get_tok_str_dbg(t2);
                        log_error(&t1->loc, "Tokens %s and %s cannot be BOR'd", s1, s2);
                        return 1;
                    }
                    dynarray_pop(tokens, NULL);
                    dynarray_pop(tokens, NULL);
                    token_t tok = {.type = TT_PUSH_INT, .int_v = val, .loc = t1->loc};
                    dynarray_push(tokens, tok);
                } else {
                    const char* s1 = get_tok_str_dbg(t1);
                    const char* s2 = get_tok_str_dbg(t2);
                    log_error(&t1->loc, "Tokens %s and %s cannot be BOR'd", s1, s2);
                    return 1;
                }
            }; break;
            case (OP_BAND): {
                token_t* t2 = &tmp_toks[dynarray_length(tmp_toks - 1)];
                token_t* t1 = &tmp_toks[dynarray_length(tmp_toks - 2)];
                if (t1->type == t2->type) {
                    size_t val = 0;
                    switch (t1->type) {
                    case (TT_PUSH_INT):
                        val = t1->int_v & t2->int_v;
                        break;
                    default:
                        const char* s1 = get_tok_str_dbg(t1);
                        const char* s2 = get_tok_str_dbg(t2);
                        log_error(&t1->loc, "Tokens %s and %s cannot be BAND'd", s1, s2);
                        return 1;
                    }
                    dynarray_pop(tokens, NULL);
                    dynarray_pop(tokens, NULL);
                    token_t tok = {.type = TT_PUSH_INT, .int_v = val, .loc = t1->loc};
                    dynarray_push(tokens, tok);
                } else {
                    const char* s1 = get_tok_str_dbg(t1);
                    const char* s2 = get_tok_str_dbg(t2);
                    log_error(&t1->loc, "Tokens %s and %s cannot be BAND'd", s1, s2);
                    return 1;
                }
            }; break;
            case (OP_NOT): {
                token_t* t1 = &tmp_toks[dynarray_length(tmp_toks - 1)];
                bool val = 0;
                switch (t1->type) {
                case (TT_PUSH_INT):
                    val = !t1->bool_v;
                    break;
                default:
                    const char* s1 = get_tok_str_dbg(t1);
                    log_error(&t1->loc, "Tokens %s cannot be negated", s1);
                    return 1;
                }
                dynarray_pop(tokens, NULL);
                dynarray_pop(tokens, NULL);
                token_t tok = {.type = TT_PUSH_BOOL, .int_v = val, .loc = t1->loc};
                dynarray_push(tokens, tok);
            }; break;
            case (OP_AND): {
                token_t* t2 = &tmp_toks[dynarray_length(tmp_toks - 1)];
                token_t* t1 = &tmp_toks[dynarray_length(tmp_toks - 2)];
                if (t1->type == t2->type) {
                    bool val = 0;
                    switch (t1->type) {
                    case (TT_PUSH_INT):
                        val = t1->bool_v && t2->bool_v;
                        break;
                    default:
                        const char* s1 = get_tok_str_dbg(t1);
                        const char* s2 = get_tok_str_dbg(t2);
                        log_error(&t1->loc, "Tokens %s and %s cannot be AND'd", s1, s2);
                        return 1;
                    }
                    dynarray_pop(tokens, NULL);
                    dynarray_pop(tokens, NULL);
                    token_t tok = {.type = TT_PUSH_BOOL, .int_v = val, .loc = t1->loc};
                    dynarray_push(tokens, tok);
                } else {
                    const char* s1 = get_tok_str_dbg(t1);
                    const char* s2 = get_tok_str_dbg(t2);
                    log_error(&t1->loc, "Tokens %s and %s cannot be SHL'd", s1, s2);
                    return 1;
                }
            }; break;
            case (OP_OR): {
                token_t* t2 = &tmp_toks[dynarray_length(tmp_toks - 1)];
                token_t* t1 = &tmp_toks[dynarray_length(tmp_toks - 2)];
                if (t1->type == t2->type) {
                    bool val = 0;
                    switch (t1->type) {
                    case (TT_PUSH_INT):
                        val = t1->bool_v || t2->bool_v;
                        break;
                    default:
                        const char* s1 = get_tok_str_dbg(t1);
                        const char* s2 = get_tok_str_dbg(t2);
                        log_error(&t1->loc, "Tokens %s and %s cannot be AND'd", s1, s2);
                        return 1;
                    }
                    dynarray_pop(tokens, NULL);
                    dynarray_pop(tokens, NULL);
                    token_t tok = {.type = TT_PUSH_BOOL, .int_v = val, .loc = t1->loc};
                    dynarray_push(tokens, tok);
                } else {
                    const char* s1 = get_tok_str_dbg(t1);
                    const char* s2 = get_tok_str_dbg(t2);
                    log_error(&t1->loc, "Tokens %s and %s cannot be SHL'd", s1, s2);
                    return 1;
                }
            }; break;
            default:
                const char* s1 = get_tok_str_dbg(&tok);
                log_error(&tok.loc, "Using %s is not allowed in Constants", s1);
                return 1;
            }
        } break;
        case (TT_NONE):
            continue;
        }
    }

    return 0;
}
